@@ -89,8 +89,8 @@ void MultilineLayout::render(cairo_t *cr, int x, int y, int lineHeight,
 
 InputWindow::InputWindow(ClassicUIConfig *config, FcitxGClient *client)
     : config_(config), client_(FCITX_G_CLIENT(g_object_ref(client))) {
-    fontMap_.reset(pango_cairo_font_map_new());
-    context_.reset(pango_font_map_create_context(fontMap_.get()));
+    context_.reset(
+        pango_font_map_create_context(pango_cairo_font_map_get_default()));
     upperLayout_ = newPangoLayout(context_.get());
     lowerLayout_ = newPangoLayout(context_.get());
 
@@ -223,10 +223,15 @@ void InputWindow::resizeCandidates(size_t n) {
 }
 void InputWindow::setLanguageAttr(size_t size, PangoAttrList *attrList,
                                   PangoAttrList *highlightAttrList) {
-    if (!config_->useInputMethodLanguageToDisplayText_ || language_.empty()) {
-        return;
-    }
-    if (auto language = pango_language_from_string(language_.c_str())) {
+    do {
+        if (!config_->useInputMethodLanguageToDisplayText_ ||
+            language_.empty()) {
+            break;
+        }
+        auto language = pango_language_from_string(language_.c_str());
+        if (!language) {
+            break;
+        }
         if (attrList) {
             auto attr = pango_attr_language_new(language);
             attr->start_index = 0;
@@ -239,7 +244,8 @@ void InputWindow::setLanguageAttr(size_t size, PangoAttrList *attrList,
             attr->end_index = size;
             pango_attr_list_insert(highlightAttrList, attr);
         }
-    }
+        return;
+    } while (0);
 }
 
 void InputWindow::setTextToMultilineLayout(MultilineLayout &layout,
@@ -357,16 +363,27 @@ void InputWindow::updateUI(GPtrArray *preedit, int cursor_pos, GPtrArray *auxUp,
     update();
 }
 
-void InputWindow::updateLanguage(const char *language) { language_ = language; }
+void InputWindow::updateLanguage(const char *language) {
+    language_ = language;
+    do {
+        if (!config_->useInputMethodLanguageToDisplayText_ ||
+            language_.empty()) {
+            break;
+        }
+        auto language = pango_language_from_string(language_.c_str());
+        if (!language) {
+            break;
+        }
+        pango_context_set_language(context_.get(), language);
+        return;
+    } while (0);
+
+    pango_context_set_language(context_.get(), pango_language_get_default());
+}
 
 std::pair<unsigned int, unsigned int> InputWindow::sizeHint() {
     auto *fontDesc = pango_font_description_from_string(config_->font_.data());
     pango_context_set_font_description(context_.get(), fontDesc);
-    if (dpi_ > 0) {
-        pango_cairo_font_map_set_resolution(
-            PANGO_CAIRO_FONT_MAP(fontMap_.get()), dpi_);
-    }
-    pango_cairo_context_set_resolution(context_.get(), dpi_);
     pango_font_description_free(fontDesc);
     pango_layout_context_changed(upperLayout_.get());
     pango_layout_context_changed(lowerLayout_.get());
@@ -462,51 +479,8 @@ void InputWindow::paint(cairo_t *cr, unsigned int width, unsigned int height) {
     const auto &textMargin = config_->theme_.textMargin;
     cairo_set_operator(cr, CAIRO_OPERATOR_OVER);
     cairo_save(cr);
-
-    prevRegion_ = cairo_rectangle_int_t{0, 0, 0, 0};
-    nextRegion_ = cairo_rectangle_int_t{0, 0, 0, 0};
-    if (nCandidates_ && (hasPrev_ || hasNext_)) {
-        const auto &prev = config_->theme_.loadAction(config_->theme_.prev);
-        const auto &next = config_->theme_.loadAction(config_->theme_.next);
-        if (prev.valid() && next.valid()) {
-            cairo_save(cr);
-            nextRegion_.x = width - margin.marginRight - next.width();
-            nextRegion_.y = height - margin.marginBottom - next.height();
-            nextRegion_.width = next.width();
-            nextRegion_.height = next.height();
-            cairo_translate(cr, nextRegion_.x, nextRegion_.y);
-            shrink(nextRegion_, config_->theme_.next.clickMargin);
-            double alpha = 1.0;
-            if (!hasNext_) {
-                alpha = 0.3;
-            } else if (nextHovered_) {
-                alpha = 0.7;
-            }
-            config_->theme_.paint(cr, config_->theme_.next, alpha);
-            cairo_restore(cr);
-            cairo_save(cr);
-            prevRegion_.x =
-                width - margin.marginRight - next.width() - prev.width();
-            prevRegion_.y = height - margin.marginBottom - prev.height();
-            prevRegion_.width = prev.width();
-            prevRegion_.height = prev.height();
-            cairo_translate(cr, prevRegion_.x, prevRegion_.y);
-            shrink(prevRegion_, config_->theme_.prev.clickMargin);
-            alpha = 1.0;
-            if (!hasPrev_) {
-                alpha = 0.3;
-            } else if (prevHovered_) {
-                alpha = 0.7;
-            }
-            config_->theme_.paint(cr, config_->theme_.prev, alpha);
-            cairo_restore(cr);
-        }
-    }
-
     // Move position to the right place.
     cairo_translate(cr, margin.marginLeft, margin.marginTop);
-
-    cairo_save(cr);
     cairoSetSourceColor(cr, config_->theme_.normalColor);
     // CLASSICUI_DEBUG() << theme.inputPanel->normalColor->toString();
     auto *metrics = pango_context_get_metrics(
@@ -636,15 +610,85 @@ void InputWindow::paint(cairo_t *cr, unsigned int width, unsigned int height) {
         }
     }
     cairo_restore(cr);
+
+    prevRegion_ = cairo_rectangle_int_t{0, 0, 0, 0};
+    nextRegion_ = cairo_rectangle_int_t{0, 0, 0, 0};
+    if (nCandidates_ && (hasPrev_ || hasNext_)) {
+        const auto &prev = config_->theme_.loadAction(config_->theme_.prev);
+        const auto &next = config_->theme_.loadAction(config_->theme_.next);
+        if (prev.valid() && next.valid()) {
+            cairo_save(cr);
+            int prevY = 0, nextY = 0;
+            switch (config_->theme_.buttonAlignment) {
+            case PageButtonAlignment::Top:
+                prevY = margin.marginTop;
+                nextY = margin.marginTop;
+                break;
+            case PageButtonAlignment::FirstCandidate:
+                prevY =
+                    candidateRegions_.front().y +
+                    (candidateRegions_.front().height - prev.height()) / 2.0;
+                nextY =
+                    candidateRegions_.front().y +
+                    (candidateRegions_.front().height - prev.height()) / 2.0;
+                break;
+            case PageButtonAlignment::Center:
+                prevY =
+                    margin.marginTop + (height - margin.marginTop -
+                                        margin.marginBottom - prev.height()) /
+                                           2.0;
+                nextY =
+                    margin.marginTop + (height - margin.marginTop -
+                                        margin.marginBottom - next.height()) /
+                                           2.0;
+                break;
+            case PageButtonAlignment::LastCandidate:
+                prevY = candidateRegions_.back().y +
+                        (candidateRegions_.back().height - prev.height()) / 2.0;
+                nextY = candidateRegions_.back().y +
+                        (candidateRegions_.back().height - next.height()) / 2.0;
+                break;
+            case PageButtonAlignment::Bottom:
+            default:
+                prevY = height - margin.marginBottom - prev.height();
+                nextY = height - margin.marginBottom - next.height();
+                break;
+            }
+            nextRegion_.x = width - margin.marginRight - next.width();
+            nextRegion_.y = nextY;
+            nextRegion_.width = next.width();
+            nextRegion_.height = next.height();
+            cairo_translate(cr, nextRegion_.x, nextRegion_.y);
+            shrink(nextRegion_, config_->theme_.next.clickMargin);
+            double alpha = 1.0;
+            if (!hasNext_) {
+                alpha = 0.3;
+            } else if (nextHovered_) {
+                alpha = 0.7;
+            }
+            config_->theme_.paint(cr, config_->theme_.next, alpha);
+            cairo_restore(cr);
+            cairo_save(cr);
+            prevRegion_.x =
+                width - margin.marginRight - next.width() - prev.width();
+            prevRegion_.y = prevY;
+            prevRegion_.width = prev.width();
+            prevRegion_.height = prev.height();
+            cairo_translate(cr, prevRegion_.x, prevRegion_.y);
+            shrink(prevRegion_, config_->theme_.prev.clickMargin);
+            alpha = 1.0;
+            if (!hasPrev_) {
+                alpha = 0.3;
+            } else if (prevHovered_) {
+                alpha = 0.7;
+            }
+            config_->theme_.paint(cr, config_->theme_.prev, alpha);
+            cairo_restore(cr);
+        }
+    }
 }
 
 void InputWindow::click(int x, int y) {
-    for (size_t idx = 0, e = candidateRegions_.size(); idx < e; idx++) {
-        if (rectContains(candidateRegions_[idx], x, y)) {
-            selectCandidate(idx);
-            return;
-        }
-    }
     if (hasPrev_ && rectContains(prevRegion_, x, y)) {
         prev();
         return;
@@ -653,6 +697,12 @@ void InputWindow::click(int x, int y) {
         next();
         return;
     }
+    for (size_t idx = 0, e = candidateRegions_.size(); idx < e; idx++) {
+        if (rectContains(candidateRegions_[idx], x, y)) {
+            selectCandidate(idx);
+            return;
+        }
+    }
 }
 
 void InputWindow::wheel(bool up) {
@@ -680,23 +730,32 @@ int InputWindow::highlight() const {
 
 bool InputWindow::hover(int x, int y) {
     bool needRepaint = false;
+
+    bool prevHovered = false;
+    bool nextHovered = false;
     auto oldHighlight = highlight();
     hoverIndex_ = -1;
-    for (int idx = 0, e = candidateRegions_.size(); idx < e; idx++) {
-        if (rectContains(candidateRegions_[idx], x, y)) {
-            hoverIndex_ = idx;
-            break;
+
+    prevHovered = rectContains(prevRegion_, x, y);
+    if (!prevHovered) {
+        nextHovered = rectContains(nextRegion_, x, y);
+        if (!nextHovered) {
+            for (int idx = 0, e = candidateRegions_.size(); idx < e; idx++) {
+                if (rectContains(candidateRegions_[idx], x, y)) {
+                    hoverIndex_ = idx;
+                    break;
+                }
+            }
         }
     }
 
-    needRepaint = needRepaint || oldHighlight != highlight();
-
-    auto prevHovered = rectContains(prevRegion_, x, y);
-    auto nextHovered = rectContains(nextRegion_, x, y);
     needRepaint = needRepaint || prevHovered_ != prevHovered;
-    needRepaint = needRepaint || nextHovered_ != nextHovered;
     prevHovered_ = prevHovered;
+
+    needRepaint = needRepaint || nextHovered_ != nextHovered;
     nextHovered_ = nextHovered;
+
+    needRepaint = needRepaint || oldHighlight != highlight();
     return needRepaint;
 }
 
